# -*- coding: utf-8 -*-

import datetime
from dateutil.parser import parse as parse_date

from .meta.datatype import DataType


class EnumType(DataType):

    def serialize(self, value):
        return value.name if value else None

    def unserialize(self, value):
        processed = self.preprocess(value)
        if isinstance(processed, self.type):
            return processed
        try:
            return self.type[processed]
        except KeyError:
            raise ValueError('Value {0} is not a member of Enum {1}'.format(
                value,
                self.type,
            ))


class ModelType(DataType):

    def serialize(self, value):
        return value.serialize() if value else None

    def unserialize(self, value):
        value = self.preprocess(value)
        return self.type(**value)


def preprocess_date(date):
    if isinstance(date, datetime.datetime):
        return date
    return parse_date(date)

datetime_type = DataType(datetime.datetime, preprocessor=preprocess_date)



# -*- coding: utf-8 -*-

import six
import decorator
import inflection
import collections
from six.moves import http_client as httplib

from . import exceptions


def get_chunks(sequence, chunk_size):
    """Split sequence into chunks.

    :param list sequence:
    :param int chunk_size:

    """
    return [
        sequence[idx:idx+chunk_size]
        for idx in range(0, len(sequence), chunk_size)
    ]


def get_kwargs(kwargs):
    """Get all keys and values from dictionary where key is not `self`.

    :param dict kwargs: Input parameters

    """
    return {
        key: value for key, value in six.iteritems(kwargs)
        if key != 'self'
    }


def check_status_code(response, codes=None):
    """Check HTTP status code and raise exception if incorrect.

    :param Response response: HTTP response
    :param codes: List of accepted codes or callable
    :raises: BetfairError if code invalid

    """
    codes = codes or [httplib.OK]
    checker = (
        codes
        if callable(codes)
        else lambda resp: resp.status_code in codes
    )
    if not checker(response):
        raise exceptions.BetfairError(response)


def result_or_error(response):
    """Get `result` field from Betfair response or raise exception if not
    found.

    :param Response response:
    :raises: BetfairAPIError if no results passed

    """
    data = response.json()
    result = data.get('result')
    if result is not None:
        return result
    raise exceptions.BetfairAPIError(response, data)


def process_result(result, model=None):
    """Cast response JSON to Betfair model(s).

    :param result: Betfair response JSON
    :param BetfairModel model: Deserialization format; if `None`, return raw
        JSON

    """
    if model is None:
        return result
    if isinstance(result, collections.Sequence):
        return [model(**item) for item in result]
    return model(**result)


def serialize_params(params):
    """Serialize input parameters to Betfair-formatted JSON.

    :param dict params: Request parameters

    """
    out = {}
    for key, value in six.iteritems(params):
        key = inflection.camelize(
            key, uppercase_first_letter=False
        ).rstrip('_')
        value = value.serialize() if hasattr(value, 'serialize') else value
        out[key] = value
    return out


def make_payload(method, params):
    """Build Betfair JSON-RPC payload.

    :param str method: Betfair endpoint
    :param dict params: Request parameters

    """
    payload = {
        'jsonrpc': '2.0',
        'method': 'SportsAPING/v1.0/{0}'.format(method),
        'params': serialize_params(params),
        'id': 1,
    }
    return payload


@decorator.decorator
def requires_login(func, *args, **kwargs):
    """Decorator to check that the user is logged in. Raises `BetfairError`
    if instance variable `session_token` is absent.

    """
    self = args[0]
    if self.session_token:
        return func(*args, **kwargs)
    raise exceptions.BetfairError('Not logged in')



# -*- coding: utf-8 -*-

import weakref
import collections

from . import exceptions


class Field(object):

    def __init__(self, data_type, required=False):
        self.data_type = data_type
        self.required = required
        self.data = weakref.WeakKeyDictionary()

    def __get__(self, instance, owner):
        try:
            return self.data[instance]
        except KeyError:
            self.data[instance] = self.missing_value()
            return self.data[instance]

    def __set__(self, instance, value, safe=False):
        value = self.data_type.unserialize(value)
        if self.required and self.is_null(value) and not safe:
            raise exceptions.MissingValueError('Value must not be `None`')
        self.data[instance] = value

    def missing_value(self):
        return None

    def is_null(self, value):
        return value is None

    def serialize(self, instance):
        value = self.__get__(instance, None)
        return self.data_type.serialize(value)


class ListContainer(collections.MutableSequence):

    def __init__(self, data_type, value=None):
        self.data = []
        self.data_type = data_type
        for item in (value or []):
            self.append(item)

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = self.data_type.unserialize(value)

    def __delitem__(self, key):
        del self.data[key]

    def __len__(self):
        return len(self.data)

    def insert(self, key, value):
        self.data.insert(key, self.data_type.unserialize(value))


class ListField(Field):

    def __set__(self, instance, value, safe=False):
        if self.required and self.is_null(value) and not safe:
            raise ValueError
        if not isinstance(value, collections.MutableSequence):
            raise ValueError
        self.data[instance] = ListContainer(self.data_type, value)

    def missing_value(self):
        return ListContainer(self.data_type)

    def is_null(self, value):
        return not value

    def serialize(self, instance):
        value = self.__get__(instance, None)
        return [
            self.data_type.serialize(item)
            for item in value
        ]



# -*- coding: utf-8 -*-

import six

from . import constants

from .meta.datatype import DataType
from .meta.field import Field, ListField
from .datatype import EnumType, ModelType, datetime_type

from .model import BetfairModel


class Event(BetfairModel):
    id = Field(DataType(six.text_type))
    name = Field(DataType(six.text_type))
    country_code = Field(DataType(six.text_type))
    timezone = Field(DataType(six.text_type))
    venue = Field(DataType(six.text_type))
    open_date = Field(datetime_type)


class MarketDescription(BetfairModel):
    persistence_enabled = Field(DataType(bool), required=True)
    bsp_market = Field(DataType(bool), required=True)
    market_time = Field(datetime_type, required=True)
    suspend_time = Field(datetime_type, required=True)
    settle_time = Field(datetime_type)
    betting_type = Field(EnumType(constants.MarketBettingType), required=True)
    turn_in_play_enabled = Field(DataType(bool), required=True)
    market_type = Field(DataType(six.text_type), required=True)
    regulator = Field(DataType(six.text_type), required=True)
    market_base_rate = Field(DataType(float), required=True)
    discount_allowed = Field(DataType(bool), required=True)
    wallet = Field(DataType(six.text_type))
    rules = Field(DataType(six.text_type))
    rules_has_date = Field(DataType(bool))
    clarifications = Field(DataType(six.text_type))


class RunnerCatalog(BetfairModel):
    selection_id = Field(DataType(six.text_type), required=True)
    runner_name = Field(DataType(six.text_type), required=True)
    handicap = Field(DataType(float), required=True)
    sort_priority = Field(DataType(int), required=True)
    metadata = Field(DataType(dict))


class EventType(BetfairModel):
    id = Field(DataType(six.text_type))
    name = Field(DataType(six.text_type))


class Competition(BetfairModel):
    id = Field(DataType(six.text_type))
    name = Field(DataType(six.text_type))


class MarketCatalogue(BetfairModel):
    market_id = Field(DataType(six.text_type))
    market_name = Field(DataType(six.text_type))
    market_start_time = Field(datetime_type)
    description = Field(EnumType(MarketDescription))
    total_matched = Field(DataType(float))
    runners = ListField(ModelType(RunnerCatalog))
    event_type = Field(ModelType(EventType))
    competition = Field(ModelType(Competition))
    event = Field(ModelType(Event))


class TimeRange(BetfairModel):
    from_ = Field(datetime_type)
    to = Field(datetime_type)


class MarketFilter(BetfairModel):
    text_query = Field(DataType(six.text_type))
    exchange_ids = ListField(DataType(six.text_type))
    event_type_ids = ListField(DataType(six.text_type))
    event_ids = ListField(DataType(six.text_type))
    competition_ids = ListField(DataType(six.text_type))
    market_ids = ListField(DataType(six.text_type))
    venues = ListField(DataType(six.text_type))
    bsp_only = Field(DataType(bool))
    turn_in_play_enabled = Field(DataType(bool))
    in_play_only = Field(DataType(bool))
    market_betting_types = ListField(EnumType(constants.MarketBettingType))
    market_countries = ListField(DataType(six.text_type))
    market_type_codes = ListField(DataType(six.text_type))
    market_start_time = Field(ModelType(TimeRange))
    with_orders = ListField(EnumType(constants.OrderStatus))


class PriceSize(BetfairModel):
    price = Field(DataType(float), required=True)
    size = Field(DataType(float), required=True)


class StartingPrices(BetfairModel):
    near_price = Field(DataType(float))
    far_price = Field(DataType(float))
    back_state_taken = Field(ModelType(PriceSize))
    lay_liability_taken = Field(ModelType(PriceSize))
    actual_SP = Field(DataType(float))


class ExchangePrices(BetfairModel):
    available_to_back = ListField(ModelType(PriceSize))
    available_to_lay = ListField(ModelType(PriceSize))
    traded_volume = ListField(ModelType(PriceSize))


class Order(BetfairModel):
    bet_id = Field(DataType(six.text_type), required=True)
    order_type = Field(EnumType(constants.OrderType), required=True)
    status = Field(EnumType(constants.OrderStatus), required=True)
    persistence_type = Field(EnumType(constants.PersistenceType), required=True)
    side = Field(EnumType(constants.Side), required=True)
    price = Field(DataType(float), required=True)
    size = Field(DataType(float), required=True)
    bsp_liability = Field(DataType(float), required=True)
    placed_date = Field(datetime_type, required=True)
    avg_price_matched = Field(DataType(float))
    size_matched = Field(DataType(float))
    size_remaining = Field(DataType(float))
    size_lapsed = Field(DataType(float))
    size_cancelled = Field(DataType(float))
    size_voided = Field(DataType(float))


class Match(BetfairModel):
    bet_id = Field(DataType(six.text_type))
    match_id = Field(DataType(six.text_type))
    side = Field(EnumType(constants.Side), required=True)
    price = Field(DataType(float), required=True)
    size = Field(DataType(float), required=True)
    match_date = Field(datetime_type)


class Runner(BetfairModel):
    selection_id = Field(DataType(float), required=True)
    handicap = Field(DataType(float), required=True)
    status = Field(EnumType(constants.RunnerStatus), required=True)
    adjustment_factor = Field(DataType(float))
    last_price_traded = Field(DataType(float))
    total_matched = Field(DataType(float))
    removal_date = Field(datetime_type)
    sp = Field(ModelType(StartingPrices))
    ex = Field(ModelType(ExchangePrices))
    orders = ListField(ModelType(Order))
    matches = ListField(ModelType(Match))


class MarketBook(BetfairModel):
    market_id = Field(DataType(six.text_type), required=True)
    is_market_data_delayed = Field(DataType(bool), required=True)
    status = Field(EnumType(constants.MarketStatus))
    bet_delay = Field(DataType(int))
    bsp_reconciled = Field(DataType(bool))
    complete = Field(DataType(bool))
    inplay = Field(DataType(bool))
    number_of_winners = Field(DataType(int))
    number_of_runners = Field(DataType(int))
    number_of_active_runners = Field(DataType(int))
    last_match_time = Field(datetime_type)
    total_matched = Field(DataType(float))
    total_available = Field(DataType(float))
    cross_matching = Field(DataType(bool))
    runners_voidable = Field(DataType(bool))
    version = Field(DataType(float))
    runners = ListField(ModelType(Runner))


class RunnerProfitAndLoss(BetfairModel):
    selection_id = Field(DataType(float))
    if_win = Field(DataType(float))
    if_lose = Field(DataType(float))


class MarketProfitAndLoss(BetfairModel):
    market_id = Field(DataType(six.text_type))
    commission_applied = Field(DataType(float))
    profit_and_losses = ListField(ModelType(RunnerProfitAndLoss))


class ExBestOffersOverrides(BetfairModel):
    best_prices_depth = Field(DataType(int))
    rollup_model = Field(EnumType(constants.RollupModel))
    rollup_limit = Field(DataType(int))
    rollup_liability_threshold = Field(DataType(float))
    rollup_liability_factor = Field(DataType(int))


class PriceProjection(BetfairModel):
    price_data = ListField(EnumType(constants.PriceData))
    ex_best_offer_overrides = Field(ModelType(ExBestOffersOverrides))
    virtualize = Field(DataType(bool))
    rollover_stakes = Field(DataType(bool))


class LimitOrder(BetfairModel):
    size = Field(DataType(float), required=True)
    price = Field(DataType(float), required=True)
    persistence_type = Field(EnumType(constants.PersistenceType), required=True)


class LimitOnCloseOrder(BetfairModel):
    liability = Field(DataType(float), required=True)
    price = Field(DataType(float), required=True)


class MarketOnCloseOrder(BetfairModel):
    liability = Field(DataType(float), required=True)


# Results

class CompetitionResult(BetfairModel):

    competition = Field(ModelType(Competition))
    market_count = Field(DataType(int))
    competition_region = Field(DataType(six.text_type))


class CountryCodeResult(BetfairModel):
    country_code = Field(DataType(six.text_type))
    market_count = Field(DataType(int))


class EventResult(BetfairModel):

    event = Field(ModelType(Event))
    market_count = Field(DataType(int))


class EventTypeResult(BetfairModel):

    event_type = Field(ModelType(EventType))
    market_count = Field(DataType(int))


class MarketTypeResult(BetfairModel):
    market_type = Field(DataType(six.text_type))
    market_count = Field(DataType(int))


class TimeRangeResult(BetfairModel):

    time_range = Field(ModelType(TimeRange))
    market_count = Field(DataType(int))


class VenueResult(BetfairModel):
    venue = Field(DataType(six.text_type))
    market_count = Field(DataType(int))


# Instructions

class PlaceInstruction(BetfairModel):
    order_type = Field(EnumType(constants.OrderType), required=True)
    selection_id = Field(DataType(float), required=True)
    handicap = Field(DataType(float))
    side = Field(EnumType(constants.Side), required=True)
    limit_order = Field(ModelType(LimitOrder))
    limit_on_close_order = Field(ModelType(LimitOnCloseOrder))
    market_on_close_order = Field(ModelType(MarketOnCloseOrder))


class CancelInstruction(BetfairModel):
    bet_id = Field(DataType(six.text_type), required=True)
    size_reduction = Field(DataType(float))


class ReplaceInstruction(BetfairModel):
    bet_id = Field(DataType(six.text_type), required=True)
    new_price = Field(DataType(float), required=True)


class UpdateInstruction(BetfairModel):
    bet_id = Field(DataType(six.text_type), required=True)
    new_persistence_type = Field(EnumType(constants.PersistenceType), required=True)


# Summary reports

class CurrentOrderSummary(BetfairModel):
    bet_id = Field(DataType(six.text_type), required=True)
    market_id = Field(DataType(six.text_type), required=True)
    selection_id = Field(DataType(float), required=True)
    handicap = Field(DataType(float), required=True)
    price_size = Field(DataType(PriceSize), required=True)
    bsp_liability = Field(DataType(float), required=True)
    side = Field(EnumType(constants.Side), required=True)
    status = Field(EnumType(constants.OrderStatus), required=True)
    persistence_type = Field(EnumType(constants.PersistenceType), required=True)
    order_type = Field(EnumType(constants.OrderType), required=True)
    placed_date = Field(datetime_type, required=True)
    matched_date = Field(datetime_type, required=True)
    average_price_matched = Field(DataType(float))
    size_matched = Field(DataType(float))
    size_remaining = Field(DataType(float))
    size_lapsed = Field(DataType(float))
    size_cancelled = Field(DataType(float))
    size_voided = Field(DataType(float))
    regulator_auth_code = Field(DataType(six.text_type))
    regulator_code = Field(DataType(six.text_type))


class CurrentOrderSummaryReport(BetfairModel):
    current_orders = ListField(ModelType(CurrentOrderSummary), required=True)
    more_available = Field(DataType(bool), required=True)


class ItemDescription(BetfairModel):
    event_type_desc = Field(DataType(six.text_type))
    event_desc = Field(DataType(six.text_type))
    market_desc = Field(DataType(six.text_type))
    market_start_Time = Field(datetime_type)
    runner_desc = Field(DataType(six.text_type))
    number_of_winners = Field(DataType(int))


class ClearedOrderSummary(BetfairModel):
    event_type_id = Field(DataType(six.text_type))
    event_id = Field(DataType(six.text_type))
    market_id = Field(DataType(six.text_type))
    selection_id = Field(DataType(six.text_type))
    handicap = Field(DataType(float))
    bet_id = Field(DataType(six.text_type))
    placed_data = Field(datetime_type)
    persistence_type = Field(EnumType(constants.PersistenceType))
    order_type = Field(EnumType(constants.OrderType))
    side = Field(EnumType(constants.Side))
    item_description = Field(ModelType(ItemDescription))
    price_requested = Field(ModelType(float))
    settled_date = Field(datetime_type)
    bet_count = Field(DataType(int))
    commission = Field(ModelType(float))
    price_matched = Field(ModelType(float))
    price_reduced = Field(DataType(bool))
    size_settled = Field(ModelType(float))
    profit = Field(ModelType(float))
    size_cancelled = Field(ModelType(float))


class ClearedOrderSummaryReport(BetfairModel):
    current_orders = ListField(ModelType(ClearedOrderSummary), required=True)
    more_available = Field(DataType(bool), required=True)


# Instruction reports

class BaseInstructionReport(BetfairModel):
    status = Field(EnumType(constants.InstructionReportStatus), required=True)
    error_code = Field(EnumType(constants.InstructionReportErrorCode))


class PlaceInstructionReport(BaseInstructionReport):
    instruction = Field(ModelType(PlaceInstruction), required=True)
    bet_id = Field(DataType(six.text_type))
    placed_date = Field(datetime_type)
    average_price_matched = Field(DataType(float))
    size_matched = Field(DataType(float))


class CancelInstructionReport(BaseInstructionReport):
    instruction = Field(ModelType(CancelInstruction))
    size_cancelled = Field(DataType(float), required=True)
    cancelled_date = Field(datetime_type)


class ReplaceInstructionReport(BaseInstructionReport):
    cancel_instruction_report = Field(ModelType(CancelInstructionReport))
    place_instruction_report = Field(ModelType(PlaceInstructionReport))


class UpdateInstructionReport(BaseInstructionReport):
    instruction = Field(ModelType(UpdateInstruction), required=True)


# Execution reports

class BaseExecutionReport(BetfairModel):
    customer_ref = Field(DataType(six.text_type))
    status = Field(EnumType(constants.ExecutionReportStatus), required=True)
    error_code = Field(EnumType(constants.ExecutionReportErrorCode))
    market_id = Field(DataType(six.text_type))


class PlaceExecutionReport(BaseExecutionReport):
    instruction_reports = ListField(ModelType(PlaceInstructionReport))


class CancelExecutionReport(BaseExecutionReport):
    instruction_reports = ListField(ModelType(CancelInstructionReport))


class ReplaceExecutionReport(BaseExecutionReport):
    instruction_reports = ListField(ModelType(ReplaceInstructionReport))


class UpdateExecutionReport(BaseExecutionReport):
    instruction_reports = ListField(ModelType(UpdateInstructionReport))



# -*- coding: utf-8 -*-

import os
import json
import requests
import itertools
from six.moves import http_client as httplib
from six.moves import urllib_parse as urllib

from . import utils
from . import models
from . import exceptions


IDENTITY_URL = 'https://identitysso.betfair.com/api/'
API_URL = 'https://api.betfair.com/exchange/betting/json-rpc/v1/'


class Betfair(object):
    """Betfair API client.

    :param str app_key: Optional application identifier
    :param str cert_file: Path to self-signed SSL certificate file(s); may be
        a *.pem file or a tuple of (*.crt, *.key) files
    :param str content_type: Response type

    """
    def __init__(self, app_key, cert_file, content_type='application/json'):
        self.app_key = app_key
        self.cert_file = cert_file
        self.content_type = content_type
        self.session = requests.Session()
        self.session_token = None

    @property
    def headers(self):
        return {
            'X-Application': self.app_key,
            'X-Authentication': self.session_token,
            'Content-Type': self.content_type,
            'Accept': 'application/json',
        }

    def make_auth_request(self, method):
        response = self.session.post(
            os.path.join(IDENTITY_URL, method),
            headers=self.headers,
        )
        utils.check_status_code(response)
        data = response.json()
        if data.get('status') != 'SUCCESS':
            raise exceptions.BetfairAuthError(response, data)

    def make_api_request(self, method, params, codes=None, model=None):
        payload = utils.make_payload(method, params)
        response = self.session.post(
            API_URL,
            data=json.dumps(payload),
            headers=self.headers,
        )
        utils.check_status_code(response, codes=codes)
        result = utils.result_or_error(response)
        return utils.process_result(result, model)

    # Authentication methods

    def login(self, username, password):
        """Log in to Betfair. Sets `session_token` if successful.

        :param str username: Username
        :param str password: Password
        :raises: BetfairLoginError

        """
        response = self.session.post(
            os.path.join(IDENTITY_URL, 'certlogin'),
            cert=self.cert_file,
            data=urllib.urlencode({
                'username': username,
                'password': password,
            }),
            headers={
                'X-Application': self.app_key,
                'Content-Type': 'application/x-www-form-urlencoded',
            },
        )
        utils.check_status_code(response, [httplib.OK])
        data = response.json()
        if data.get('loginStatus') != 'SUCCESS':
            raise exceptions.BetfairLoginError(response, data)
        self.session_token = data['sessionToken']

    @utils.requires_login
    def keep_alive(self):
        """Reset session timeout.

        :raises: BetfairAuthError

        """
        self.make_auth_request('keepAlive')

    @utils.requires_login
    def logout(self):
        """Log out and clear `session_token`.

        :raises: BetfairAuthError

        """
        self.make_auth_request('logout')
        self.session_token = None

    # Bet query methods

    @utils.requires_login
    def list_event_types(self, filter, locale=None):
        """

        :param MarketFilter filter:
        :param str locale:

        """
        return self.make_api_request(
            'listEventTypes',
            utils.get_kwargs(locals()),
            model=models.EventTypeResult,
        )

    @utils.requires_login
    def list_competitions(self, filter, locale=None):
        """

        :param MarketFilter filter:
        :param str locale:

        """
        return self.make_api_request(
            'listCompetitions',
            utils.get_kwargs(locals()),
            model=models.CompetitionResult,
        )

    @utils.requires_login
    def list_time_ranges(self, filter, granularity):
        """

        :param MarketFilter filter:
        :param TimeGranularity granularity:

        """
        return self.make_api_request(
            'listTimeRanges',
            utils.get_kwargs(locals()),
            model=models.TimeRangeResult,
        )

    @utils.requires_login
    def list_events(self, filter, locale=None):
        """

        :param MarketFilter filter:
        :param str locale:

        """
        return self.make_api_request(
            'listEvents',
            utils.get_kwargs(locals()),
            model=models.EventResult,
        )

    @utils.requires_login
    def list_market_types(self, filter, locale=None):
        """

        :param MarketFilter filter:
        :param str locale:

        """
        return self.make_api_request(
            'listMarketTypes',
            utils.get_kwargs(locals()),
            model=models.MarketTypeResult,
        )

    @utils.requires_login
    def list_countries(self, filter, locale=None):
        """

        :param MarketFilter filter:
        :param str locale:

        """
        return self.make_api_request(
            'listCountries',
            utils.get_kwargs(locals()),
            model=models.CountryCodeResult,
        )

    @utils.requires_login
    def list_venues(self, filter, locale=None):
        """

        :param MarketFilter filter:
        :param str locale:

        """
        return self.make_api_request(
            'listCountries',
            utils.get_kwargs(locals()),
            model=models.VenueResult,
        )

    @utils.requires_login
    def list_market_catalogue(
            self, filter, max_results=100, market_projection=None, locale=None,
            sort=None):
        """

        :param MarketFilter filter:
        :param int max_results:
        :param list market_projection:
        :param MarketSort sort:
        :param str locale:

        """
        return self.make_api_request(
            'listMarketCatalogue',
            utils.get_kwargs(locals()),
            model=models.MarketCatalogue,
        )

    @utils.requires_login
    def list_market_book(
            self, market_ids, price_projection=None, order_projection=None,
            match_projection=None, currency_code=None, locale=None):
        """

        :param list market_ids: List of market IDs
        :param PriceProjection price_projection:
        :param OrderProjection order_projection:
        :param MatchProjection match_projection:
        :param str currency_code:
        :param str locale:

        """
        return self.make_api_request(
            'listMarketBook',
            utils.get_kwargs(locals()),
            model=models.MarketBook,
        )

    @utils.requires_login
    def list_market_profit_and_loss(
            self, market_ids, include_settled_bets=False,
            include_bsp_bets=None, net_of_commission=None):
        """Retrieve profit and loss for a given list of markets.

        :param list market_ids: List of markets to calculate profit and loss
        :param bool include_settled_bets: Option to include settled bets
        :param bool include_bsp_bets: Option to include BSP bets
        :param bool net_of_commission: Option to return profit and loss net of
            users current commission rate for this market including any special
            tariffs

        """
        return self.make_api_request(
            'listMarketProfitAndLoss',
            utils.get_kwargs(locals()),
            model=models.MarketProfitAndLoss,
        )

    # Chunked iterators for list methods

    def iter_list_market_book(self, market_ids, chunk_size, **kwargs):
        """Split call to `list_market_book` into separate requests.

        :param list market_ids: List of market IDs
        :param int chunk_size: Number of records per chunk
        :param dict kwargs: Arguments passed to `list_market_book`

        """
        return itertools.chain(*(
            self.list_market_book(market_chunk, **kwargs)
            for market_chunk in utils.get_chunks(market_ids, chunk_size)
        ))

    def iter_list_market_profit_and_loss(
            self, market_ids, chunk_size, **kwargs):
        """Split call to `list_market_profit_and_loss` into separate requests.

        :param list market_ids: List of market IDs
        :param int chunk_size: Number of records per chunk
        :param dict kwargs: Arguments passed to `list_market_profit_and_loss`

        """
        return itertools.chain(*(
            self.list_market_profit_and_loss(market_chunk, **kwargs)
            for market_chunk in utils.get_chunks(market_ids, chunk_size)
        ))

    # Betting methods

    @utils.requires_login
    def list_current_orders(
            self, bet_ids, market_ids, order_projection, date_range, order_by,
            sort_dir, from_record, record_count):
        """

        :param bet_ids:
        :param market_ids:
        :param order_projection:
        :param date_range:
        :param order_by:
        :param sort_dir:
        :param from_record:
        :param record_count:

        """
        return self.make_api_request(
            'listCurrentOrders',
            utils.get_kwargs(locals()),
            model=models.CurrentOrderSummaryReport,
        )

    @utils.requires_login
    def list_cleared_orders(
            self, bet_status, event_type_ids, event_ids, market_ids,
            runner_ids, bet_ids, side, settled_date_range, group_by,
            include_item_description, locale, from_record, record_count):
        """

        :param bet_status:
        :param event_type_ids:
        :param event_ids:
        :param market_ids:
        :param runner_ids:
        :param bet_ids:
        :param side:
        :param settled_date_range:
        :param group_by:
        :param include_item_description:
        :param locale:
        :param from_record:
        :param record_count:

        """
        return self.make_api_request(
            'listClearedOrders',
            utils.get_kwargs(locals()),
            model=models.ClearedOrderSummaryReport,
        )

    @utils.requires_login
    def place_orders(self, market_id, instructions, customer_ref=None):
        """Place new orders into market. This operation is atomic in that all
        orders will be placed or none will be placed.

        :param str market_id: The market id these orders are to be placed on
        :param list instructions: List of `PlaceInstruction` objects
        :param str customer_ref: Optional order identifier string

        """
        return self.make_api_request(
            'placeOrders',
            utils.get_kwargs(locals()),
            model=models.PlaceExecutionReport,
        )

    @utils.requires_login
    def cancel_orders(self, market_id, instructions, customer_ref=None):
        """Cancel all bets OR cancel all bets on a market OR fully or
        partially cancel particular orders on a market.

        :param str market_id: If not supplied all bets are cancelled
        :param list instructions: List of `CancelInstruction` objects
        :param str customer_ref: Optional order identifier string

        """
        return self.make_api_request(
            'cancelOrders',
            utils.get_kwargs(locals()),
            model=models.CancelInstructionReport,
        )

    @utils.requires_login
    def replace_orders(self, market_id, instructions, customer_ref=None):
        """This operation is logically a bulk cancel followed by a bulk place.
        The cancel is completed first then the new orders are placed.

        :param str market_id: The market id these orders are to be placed on
        :param list instructions: List of `ReplaceInstruction` objects
        :param str customer_ref: Optional order identifier string

        """
        return self.make_api_request(
            'replaceOrders',
            utils.get_kwargs(locals()),
            model=models.ReplaceExecutionReport,
        )

    @utils.requires_login
    def update_orders(self, market_id, instructions, customer_ref=None):
        """Update non-exposure changing fields.

        :param str market_id: The market id these orders are to be placed on
        :param list instructions: List of `UpdateInstruction` objects
        :param str customer_ref: Optional order identifier string

        """
        return self.make_api_request(
            'updateOrders',
            utils.get_kwargs(locals()),
            model=models.UpdateExecutionReport,
        )




# -*- coding: utf-8 -*-

import six
import copy

from .field import Field


class Model(object):
    """Model base class. Subclasses should define `Field` descriptors, which
    are collected in `_fields`. Note: to avoid circular definitions, the
    `ModelMeta` metaclass is not applied until after `Model` has been defined.

    """
    def __init__(self, **kwargs):
        self.unserialize(kwargs)

    def serialize_key(self, key):
        return key

    def unserialize_key(self, key):
        return key

    def serialize(self):
        return {
            self.serialize_key(key): value.serialize(self)
            for key, value in six.iteritems(self._fields)
        }
        
    def unserialize(self, kwargs):
        for key, value in six.iteritems(kwargs):
            key = self.unserialize_key(key)
            if key not in self._fields:
                raise ValueError('Key {0} not in model schema'.format(key))
            self._fields[key].__set__(self, value, safe=True)
        self.check_complete()

    def check_complete(self):
        missing = [
            key for key, value in six.iteritems(self._fields)
            if value.required
            and value.is_null(value.__get__(self, self.__class__))
        ]
        if missing:
            raise ValueError('Missing values on fields {0}'.format(
                ', '.join(missing)
            ))


def copy_fields(bases):
    """Deep-copy and collect `Field` descriptors from base classes.

    :param list bases: Base classes
    :returns: Dictionary of `Field` descriptors

    """
    fields = {}
    for base in bases[::-1]:
        if issubclass(base, Model):
            fields.update({
                key: copy.deepcopy(value)
                for key, value in six.iteritems(base._fields)
            })
    return fields


class ModelMeta(type):
    """Model metaclass. Collects field descriptors and inherits fields from
    parent.

    """
    def __init__(cls, name, bases, dct):
        cls._fields = copy_fields(bases)
        cls._fields.update({
            key: value
            for key, value in six.iteritems(dct)
            if isinstance(value, Field)
        })
        super(ModelMeta, cls).__init__(name, bases, dct)


# Apply metaclass to `Model`
Model = six.add_metaclass(ModelMeta)(Model)



# -*- coding: utf-8 -*-


class DataType(object):

    def __init__(self, type, preprocessor=None):
        self.type = type
        self.preprocessor = preprocessor

    def preprocess(self, value):
        return self.preprocessor(value) if self.preprocessor else value

    def serialize(self, value):
        return value

    def unserialize(self, value):
        processed = self.preprocess(value)
        if isinstance(processed, self.type):
            return processed
        return self.type(processed)



# -*- coding: utf-8 -*-

import inflection

from .meta.model import Model


class BetfairModel(Model):
    """Handle conversions between internal (underscore) and external (camel-
    cased) key formatting; handle trailing underscores used to avoid conflicts
    with build-ins (e.g. from_).

    """
    def serialize_key(self, key):
        return inflection.camelize(
            key, uppercase_first_letter=False
        ).rstrip('_')

    def unserialize_key(self, key):
        key = inflection.underscore(key)
        if key in self._fields:
            return key
        return key + '_'
